/* RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
 * Copyright (C) 2009-2010  The RevKit Developers <revkit@informatik.uni-bremen.de>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "costs.hpp"

#include <boost/tuple/tuple.hpp>

#include "../target_tags.hpp"

namespace revkit
{

  unsigned gate_costs::operator()( const circuit& circ ) const
  {
    return circ.num_gates();
  }

  unsigned line_costs::operator()( const circuit& circ ) const
  {
    return circ.lines();
  }

  quantum_costs::quantum_costs()
    : controls_offset( 0 )
  {
  }

  unsigned quantum_costs::operator()( const gate& g, unsigned lines ) const
  {
    unsigned costs = 0u;

    unsigned n = lines;
    unsigned c = std::distance( g.begin_controls(), g.end_controls() );

    if ( is_fredkin( g ) )
    {
      costs = 2u;
      c += 1u;
    }
    
    c += std::max( -1 * (int)c, controls_offset );
    c = std::min( c, lines - 1u );
    
    unsigned e = n - c - 1u; // empty lines

    switch ( c ) {
    case 0u:
    case 1u:
      costs = 1u;
      break;
    case 2u:
      costs = 5u;
      break;
    case 3u:
      costs = 13u;
      break;
    case 4u:
      costs = ( e >= 2u ) ? 26u : 29u;
      break;
    case 5u:
      if ( e >= 3u ) {
        costs = 38u;
      } else if ( e >= 1u ) {
        costs = 52u;
      } else {
        costs = 61u;
      }
      break;
    case 6u:
      if ( e >= 4u ) {
        costs = 50u;
      } else if ( e >= 1u ) {
        costs = 80u;
      } else {
        costs = 125u;
      }
      break;
    case 7u:
      if ( e >= 5u ) {
        costs = 62u;
      } else if ( e >= 1u ) {
        costs = 100u;
      } else {
        costs = 253u;
      }
      break;
    case 8u:
      if ( e >= 6u ) {
        costs = 74u;
      } else if ( e >= 1u ) {
        costs = 128u;
      } else {
        costs = 509u;
      }
      break;
    case 9u:
      if ( e >= 7u ) {
        costs = 86u;
      } else if ( e >= 1u ) {
        costs = 152u;
      } else {
        costs = 1021u;
      }
      break;
    default:
      if ( e >= c - 2u ) {
        costs = 12u * c - 33u;
      } else if ( e >= 1u ) {
        costs = 24u * c - 87u;
      } else {
        costs = ( 1u << ( c + 1u ) ) - 3u;
      }
    }

    return costs;
  }

  unsigned transistor_costs::operator()( const gate& g, unsigned lines ) const
  {
    return 8 * std::distance( g.begin_controls(), g.end_controls() );
  }

  struct costs_visitor : public boost::static_visitor<unsigned>
  {
    explicit costs_visitor( const circuit& circ ) : circ( circ ) {}

    unsigned operator()( const costs_by_circuit_func& f ) const
    {
      return f( circ );
    }

    unsigned operator()( const costs_by_gate_func& f ) const
    {
      unsigned sum = 0;
      foreach ( const gate& g, circ )
      {
        sum += f( g, circ.lines() );
      }
      return sum;
    }

  private:
    const circuit& circ;
  };

  unsigned costs( const circuit& circ, const cost_function& f )
  {
    return boost::apply_visitor( costs_visitor( circ ), f );
  }

}
